/******************************************************************************
* The MIT License
*
* Copyright (c) 2017 Paul Young
*
* Permission is hereby granted, free of charge, to any person
* obtaining a copy of this software and associated documentation
* files (the "Software"), to deal in the Software without
* restriction, including without limitation the rights to use, copy,
* modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
* BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
* ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*****************************************************************************/


#include "stdafx.h"
#include "Keyer.h"

// This code was designed to support the ten dollar keyer and also to
// support (with limitations) a K1EL Winkey.  The Winkey support is not
// fully implemented.  In particular the sidetone handline is different
// for a Winkey and there is no way to read a Winkey's settings for things
// such as farnsworth.

static const unsigned char INITIAL_MESSAGE[] = {
	0, 3,										// Close interface
	0, 4, '1', 0, 4, '0', 0, 4, '$', 0, 4, 'K',	// Echo 10$K
	0, 2										// Open interface	
};

static const int INITIAL_MESSAGE_SIZE = sizeof(INITIAL_MESSAGE);

static const char INITIAL_RECEIVED_MESSAGE[] = {
	'1', '0', '$'
};

static const char INITIAL_RECEIVED_MESSAGE_SIZE = sizeof(INITIAL_RECEIVED_MESSAGE);

static const unsigned char STATUS_COMMAND[1] = { 0x15 };
static const int STATUS_COMMAND_SIZE = 1;

static const unsigned char SPEED_REQUEST[] = { 
	0x05, 5, 45, 0xff,				// "Pot" speed range 5 - 50 WPM
	0x07							// Get keyer speed
};

static const int SPEED_REQUEST_SIZE = sizeof(SPEED_REQUEST);

// This requests various ten dollar keyer parameters.  For a
// WinKey the only value available is the keyer speed.
static const unsigned char FULL_REQUEST[] = {
	0x73, 0x01,						// Turn on control change notification
	0x05, 5, 45, 0xff,				// "Pot" speed range 5 - 50 WPM
	0x71, 'F', 'R', 'E', 'Q',		// Get sidetone frequency
	0x71, 'U', 'P', 'H', 'O',		// Get use speaker or headphones
	0x71, 'S', 'V', 'O', 'L',		// Get speaker volume
	0x71, 'H', 'V', 'O', 'L',		// Get headphones volume
	0x71, 'K', 'F', 'R', 'N',		// Get farnsworth speed
	0x71, 'K', 'T', 'Y', 'P',		// Get keyer type
	0x07							// Get keyer speed
};

static const int FULL_REQUEST_SIZE = sizeof(FULL_REQUEST);

static const unsigned char SHUTDOWN_MESSAGE[] = {
	0, 3										// Close interface
};

static const int SHUTDOWN_MESSAGE_SIZE = sizeof(SHUTDOWN_MESSAGE);

CKeyer::CKeyer()
{
	m_keyer_state = KEYER_STATE_DISCONNECTED;
	m_event_handle = NULL;
	m_paddle = false;
	m_speaker_volume = -1;
	m_headphones_volume = -1;
	m_keyer_type_sent = false;
	m_exiting = false;
	m_keyer_type = KEYER_TYPE_UNKNOWN;
}

CKeyer::~CKeyer()
{
	send(SHUTDOWN_MESSAGE, SHUTDOWN_MESSAGE_SIZE);
	m_exiting = true;
	disconnect();
}

void
CKeyer::set_port(int port)
//----------------------------------------------------------------------
// Set the keyer COM (serial) port.  This will disconnect the keyer.
//----------------------------------------------------------------------
{
	if (m_port != port) {
		m_port = port;
		if (m_keyer_state != KEYER_STATE_DISCONNECTED) {
			disconnect();
		}
	}
}

void
CKeyer::send(const unsigned char *data, int length)
//----------------------------------------------------------------------
// Send a buffer to the keyer
//----------------------------------------------------------------------
{
	if (m_keyer_state == KEYER_STATE_OPERATE) {
		send_internal(data, length);
	}
}

void
CKeyer::set_frequency(int frequency)
//----------------------------------------------------------------------
// Set the sidetone frequency (in 10's of Hz) = 10 dollar keyer only
//----------------------------------------------------------------------
{
	unsigned char buf[6] = { 0x72, 'F', 'R', 'E', 'Q', 0 };
	buf[5] = frequency;
	send(buf, 6);
}

void
CKeyer::set_volume(int volume)
//----------------------------------------------------------------------
// Set the sidetone volume (in dB) - 10 dollar keyer only
//----------------------------------------------------------------------
{
	unsigned char buf[6] = { 0x72, 'S', 'V', 'O', 'L', 0 };
	if (m_use_headphones) {
		buf[1] = 'H';
	}
	buf[5] = volume;
	send(buf, 6);
}

void
CKeyer::set_speed(int speed)
//----------------------------------------------------------------------
// Set the keyer speed (WPM)
//----------------------------------------------------------------------
{
	unsigned char buf[2] = { 0x2, 0 };
	buf[1] = speed;
	send(buf, 2);
}

void
CKeyer::set_farnsworth(int farnsworth)
//----------------------------------------------------------------------
// Set the Farnsworth speed (WPM)
//----------------------------------------------------------------------
{
	unsigned char buf[2] = { 0x0D, 0 };
	buf[1] = farnsworth;
	send(buf, 2);
}

void
CKeyer::set_type(Keyer_Type type)
//----------------------------------------------------------------------
// Set the keyer type
//----------------------------------------------------------------------
{
	m_keyer_type = type;
	if (m_keyer_state == KEYER_STATE_OPERATE) {
		set_type_internal(type);
	}

	return;
}

void
CKeyer::set_type_internal(Keyer_Type type)
//----------------------------------------------------------------------
// Helper routine for setting the keyer type
//----------------------------------------------------------------------
{
	// The keyer type uses the same command as setting
	// character echo for computer and paddle.  This
	// routine has the side effect of turning on echo.
	unsigned char buf[6] = { 0x0E, 0, 0x09, 0 };

	switch (type) {
	case KEYER_TYPE_IAMBIC_A:
		buf[1] = 54;
		buf[3] = 0x06;
		break;

	case KEYER_TYPE_IAMBIC_B:
		buf[1] = 0x44;
		buf[3] = 0x06;
		break;

	case KEYER_TYPE_ULTIMATIC:
		buf[1] = 0x64;
		buf[3] = 0x06;
		break;

	case KEYER_TYPE_DIT:
		buf[1] = 0x64;
		buf[3] = 0x86;
		break;

	case KEYER_TYPE_DAH:
		buf[1] = 0x64;
		buf[3] = 0x46;
		break;
	}

	send(buf, 4);
	m_keyer_type_sent = true;
}


bool
CKeyer::send_internal(const unsigned char *data, int length)
//----------------------------------------------------------------------
// Helper routine for sending a buffer to the keyer
//----------------------------------------------------------------------
{
	OVERLAPPED write_overlapped = { 0 };
	DWORD written;
	bool success = true;

	// Create this writes OVERLAPPED structure hEvent.
	write_overlapped.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (write_overlapped.hEvent == NULL) {
		disconnect();
		return false;
	}

	// Issue write.
	if (!WriteFile(m_serial_handle, data, length, &written, &write_overlapped)) {
		if (GetLastError() != ERROR_IO_PENDING) {
			// WriteFile failed, but it isn't delayed. Report error and abort.
			disconnect();
			success = false;
		}
		else {
			// Write is pending.
			WaitForSingleObject(write_overlapped.hEvent, INFINITE);

			if (!GetOverlappedResult(m_serial_handle, &write_overlapped, &written, TRUE)) {
				disconnect();
				success = false;
			}
		}
	}

	CloseHandle(write_overlapped.hEvent);
	return success;
}

void
CKeyer::disconnect()
//----------------------------------------------------------------------
// Disconnect from the keyer, close the serial connection
//----------------------------------------------------------------------
{
	if (m_keyer_state != KEYER_STATE_DISCONNECTED) {
		m_keyer_state = KEYER_STATE_DISCONNECTED;
		if (!m_exiting) {
			CloseHandle(m_serial_handle);
			SetEvent(m_event_handle);
			::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_DISCONNECT, 0);
		}
	}
}

DWORD
CKeyer::connect()
//----------------------------------------------------------------------
// Connect to the keyer using the serial port
//----------------------------------------------------------------------
{
	CString port;

	port.Format(L"\\\\.\\COM%d", m_port);

	// Open the device
	m_serial_handle = CreateFile(port,
		GENERIC_READ | GENERIC_WRITE,
		0,
		(LPSECURITY_ATTRIBUTES)NULL,
		OPEN_EXISTING,
		FILE_FLAG_OVERLAPPED,
		NULL);

	if (m_serial_handle == INVALID_HANDLE_VALUE) {
		return GetLastError();
	}


	if (m_event_handle == NULL) {
		m_event_handle = CreateEvent(NULL, TRUE, TRUE, L"");
		if (m_event_handle == NULL) {
			DWORD error = GetLastError();
			CloseHandle(m_serial_handle);
			return error;
		}
	}

	//Set the members of the overlapped structure.
	m_serial_overlapped.hEvent = m_event_handle;
	m_serial_overlapped.Offset = 0;
	m_serial_overlapped.OffsetHigh = 0;
	
	ResetEvent(m_event_handle);
	m_timeout = 0;
	m_parse_buf_len = 0;
	return ERROR_SUCCESS;
}

bool
CKeyer::receive()
//----------------------------------------------------------------------
// Receive data from the keyer
//----------------------------------------------------------------------
{
	DWORD result;
	DWORD bytes;
	unsigned char  data;
	DWORD bytes_read;

	ResetEvent(m_event_handle);
	if (m_keyer_state == KEYER_STATE_DISCONNECTED) {
		return false;
	}

	ReadFile(m_serial_handle, &data, 1, &bytes_read, (LPOVERLAPPED)&m_serial_overlapped);

	// IO pending is expected
	result = GetLastError();
	if (result == ERROR_IO_PENDING) {

		// Wait for the IO to complete or the main loop do wake us because
		// it disconnected

		result = WaitForSingleObject(m_event_handle, 1000000);
		// Check for timeout
		if (result == WAIT_TIMEOUT) {
			CancelIo(m_serial_handle);
			m_timeout++;
			if (m_timeout >= 5) {
				return false;
			}
			send_internal(STATUS_COMMAND, STATUS_COMMAND_SIZE);
			return true;
		}

		if (m_keyer_state == KEYER_STATE_DISCONNECTED) {
			return false;
		}

		if (GetOverlappedResult(m_serial_handle,
			(LPOVERLAPPED)&m_serial_overlapped, &bytes, TRUE)) {
			// Post the character if one was read
			if (bytes) {
				parse(data);
			}
			m_timeout = 0;
			return true;
		}
		return false;
	}
	else {
		// There may have already been characters, no wait needed
		if (result == 0) {
			// Post the character if one was read
			if (bytes_read) {
				parse(data);
			}
			m_timeout = 0;
			return true;
		}
		return false;
	}
}

void
CKeyer::parse(unsigned char c)
//----------------------------------------------------------------------
// Parse data received from the keyer
//----------------------------------------------------------------------
{
	if (m_keyer_state == KEYER_STATE_INITIAL) {
		// There might be characters sent by the keyer in the
		// computer's serial port buffer.  Sync by looking for
		// the 10$ string which was echoed.
		if (m_parse_buf_len < INITIAL_RECEIVED_MESSAGE_SIZE) {
			if (c != INITIAL_RECEIVED_MESSAGE[m_parse_buf_len]) {
				m_parse_buf_len = 0;
				return;
			}
		}

		// The ten dollar keyer has a hack.  If the specific
		// string 10$K is sent it will echo with a lower case k.
		// This is used to tell that it isn't a WinKey.
		m_parse_buf[m_parse_buf_len++] = c;
		if (m_parse_buf_len == 5) {
			if (m_parse_buf[3] == 'k') {
				m_keyer_model = KEYER_MODEL_TEN_DOLLAR;
				send_internal(FULL_REQUEST, FULL_REQUEST_SIZE);
			}
			else {
				// The pitch selections are different for WinKey 1
				// and WinKey 2.  
				if (m_parse_buf[4] = 0x0a) {
					m_keyer_model = KEYER_MODEL_WK1;
				}
				else {
					m_keyer_model = KEYER_MODEL_WK2;
				}
				send_internal(SPEED_REQUEST, SPEED_REQUEST_SIZE);
			}

			m_keyer_state = KEYER_STATE_OPERATE;
			::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_CONNECT, 0);
			m_parse_buf_len = 0;

			if (m_keyer_model != KEYER_MODEL_TEN_DOLLAR) {
				// The WinKey protocol has no way to query the keyer type.  So
				// if it isn't a Ten Dollar keyer the current type is unknown.
				// If nothing was specified in the Setup dialog set it to the
				// default.  This is necessary because setting the keyer to echo
				// characters requires setting the keyer type.
				if (m_keyer_type == KEYER_TYPE_UNKNOWN) {
					m_keyer_type = KEYER_TYPE_DEFAULT;
				}
				::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_TYPE, m_keyer_type);
				set_type_internal(m_keyer_type);
			}
		}
		return;
	}

	if (m_keyer_state == KEYER_STATE_OPERATE) {
		if (m_parse_buf_len) {
			m_parse_buf[m_parse_buf_len++] = c;
			if (m_parse_buf_len < 6) {
				return;
			}

			if (!memcmp(m_parse_buf + 1, "FREQ", 4)) {
				::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_FREQUENCY, (LPARAM)c * 10);
				m_parse_buf_len = 0;
				return;
			}

			if (!memcmp(m_parse_buf + 1, "UPHO", 4)) {
				if (c) {
					m_use_headphones = true;
					if (m_headphones_volume >= 0) {
						::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_VOLUME, m_headphones_volume);
					}
				}
				else {
					m_use_headphones = false;
					if (m_speaker_volume >= 0) {
						::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_VOLUME, m_speaker_volume);
					}
				}
				m_parse_buf_len = 0;
				return;
			}

			if (!memcmp(m_parse_buf + 1, "SVOL", 4)) {
				m_speaker_volume = c;
				if (!m_use_headphones) {
					::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_VOLUME, m_speaker_volume);
				}
				m_parse_buf_len = 0;
				return;
			}

			if (!memcmp(m_parse_buf + 1, "HVOL", 4)) {
				m_headphones_volume = c;
				if (m_use_headphones) {
					::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_VOLUME, m_headphones_volume);
				}
				m_parse_buf_len = 0;
				return;
			}

			if (!memcmp(m_parse_buf + 1, "KFRN", 4)) {
				::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_FARNSWORTH, c);
				m_parse_buf_len = 0;
				return;
			}

			if (!memcmp(m_parse_buf + 1, "KTYP", 4)) {
				// The first time after a connection is made the keyer type
				// is sent as the result of a query.  If a value was set by
				// the setup dialog use that instead.  Any other time is caused
				// by the user changing the setting on the keyer so use that.
				if (m_keyer_type_sent) {
					m_keyer_type = (Keyer_Type)c;
					::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_TYPE, c);
				}
				else {
					if (m_keyer_type == KEYER_TYPE_UNKNOWN) {
						m_keyer_type = (Keyer_Type)c;
						::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_TYPE, c);
					}
					// It is necessary to set the keyer type after a connection
					// because it has the side effect of turning on character echo.
					set_type_internal(m_keyer_type);
				}
				m_parse_buf_len = 0;
				return;
			}

			// There are other config parameters which may be sent.  Ignore them.
			m_parse_buf_len = 0;
			return;
		}

		// 0x71 is the start of a config parameter
		if (c == 0x71) {
			m_parse_buf[m_parse_buf_len++] = c;
			return;
		}

		// 0xC0 or greater is a status indicator
		if ((c & 0xc0) == 0xc0) {
			if (c & 4) {
				// Sending from the paddle
				m_paddle = true;
			}
			else {
				m_paddle = false;
			}
			return;
		}

		// 0x80 through 0xbf is speed
		if ((c & 0xc0) == 0x80) {
			::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_SPEED, (c & 0x7F) + MIN_SPEED);
			return;
		}

		::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_CHAR, c);
	}
}

void
CKeyer::thread_loop()
//----------------------------------------------------------------------
// Loop - receive keyer data, parse, send to main thread
//----------------------------------------------------------------------
{
	while (1) {
		if (m_keyer_state == KEYER_STATE_DISCONNECTED) {
			if (!m_port) {
				Sleep(1000);
				continue;
			}

			DWORD status = connect();
			if (status != ERROR_SUCCESS) {
				::PostMessage(m_hwnd, KEYER_MSG, KEYER_MESSAGE_DISCONNECT, status);
				Sleep(1000);
				continue;
			}

			// Send the initial message
			if (!send_internal(INITIAL_MESSAGE, INITIAL_MESSAGE_SIZE)) {
				Sleep(1000);
				continue;
			}

			m_keyer_state = KEYER_STATE_INITIAL;
		}

		if (!receive()) {
			disconnect();
		}
	}
}

void
CKeyer::start(HWND hwnd)
//----------------------------------------------------------------------
// Start receive thread
//----------------------------------------------------------------------
{
	m_hwnd = hwnd;
	AfxBeginThread(trampoline, this);
}

UINT
CKeyer::trampoline(LPVOID param)
//----------------------------------------------------------------------
// Enter thread loop from thread start
//----------------------------------------------------------------------
{
	CKeyer *keyer = (CKeyer *)param;
	keyer->thread_loop();
	return 0;
}